@@ -1,3 +1,6 @@
+"""
+Primary symbols include classes and functions for visualization, configuration, timing and generally useful utilities.
+"""
 from . import models
 from . import ops
 from . import util
@@ -7,4 +10,5 @@ from . import mpi
 from .data import toydata
 from .data.misc import universal_dict_collate_fn, to_tensor
 from .util import *
+from .visualization import *
 from .__meta__ import __version__
