@@ -2,7 +2,8 @@ import numpy as np
 import inspect
 import torch
 import torch.nn as nn
-from typing import Union, List, Tuple, Any, Dict as TDict, Iterator
+import torch.nn.init as init
+from typing import Union, List, Tuple, Any, Dict as TDict, Iterator, Type
 from torch import Tensor
 from torch.hub import load_state_dict_from_url
 import hashlib
@@ -12,12 +13,14 @@ from os.path import join
 from os import makedirs
 import pynvml as nv
 from cv2 import getGaussianKernel
+import h5py
 
-__all__ = ['Dict', 'lookup_nn', 'reduce_loss_dict', 'to_device', 'asnumpy', 'fetch_model', 'random_code_name',
-           'dict_hash', 'fetch_image', 'random_seed', 'tweak_module_', 'add_to_loss_dict',
+__all__ = ['Dict', 'lookup_nn', 'reduce_loss_dict', 'tensor_to', 'to_device', 'asnumpy', 'fetch_model',
+           'random_code_name', 'dict_hash', 'fetch_image', 'random_seed', 'tweak_module_', 'add_to_loss_dict',
            'random_code_name_dir', 'get_device', 'num_params', 'count_submodules', 'train_epoch', 'Bytes', 'Percent',
            'GpuStats', 'trainable_params', 'frozen_params', 'Tiling', 'load_image', 'gaussian_kernel',
-           'replace_module_']
+           'iter_submodules', 'replace_module_', 'wrap_module_', 'spectral_norm_', 'to_h5', 'to_tiff',
+           'exponential_moving_average_', 'from_json', 'to_json']
 
 
 class Dict(dict):
@@ -105,6 +108,33 @@ def add_to_loss_dict(d: dict, key: str, loss: torch.Tensor, weight=None):
     d[key] = loss if dk is None else dk + loss
 
 
+def tensor_to(inputs: Union[list, tuple, dict, Tensor], *args, **kwargs):
+    """Tensor to device/dtype/other.
+
+    Recursively calls ``tensor.to(*args, **kwargs)`` for all ``Tensors`` in ``inputs``.
+
+    Notes:
+        - Works recursively.
+        - Non-Tensor items are not altered.
+
+    Args:
+        inputs: Tensor, list, tuple or dict. Non-Tensor objects are ignored. Tensors are substituted by result of
+            ``tensor.to(*args, **kwargs)`` call.
+        *args: Arguments. See docstring of ``torch.Tensor.to``.
+        **kwargs: Keyword arguments. See docstring of ``torch.Tensor.to``.
+
+    Returns:
+        Inputs with Tensors replaced by ``tensor.to(*args, **kwargs)``.
+    """
+    if isinstance(inputs, Tensor):
+        inputs = inputs.to(*args, **kwargs)
+    elif isinstance(inputs, dict):
+        inputs = {k: tensor_to(b, *args, **kwargs) for k, b in inputs.items()}
+    elif isinstance(inputs, (list, tuple)):
+        inputs = type(inputs)([tensor_to(b, *args, **kwargs) for b in inputs])
+    return inputs
+
+
 def to_device(batch: Union[list, tuple, dict, Tensor], device):
     """To device.
 
@@ -116,19 +146,13 @@ def to_device(batch: Union[list, tuple, dict, Tensor], device):
         - Non-Tensor items are not altered.
 
     Args:
-        batch: Input.
+        batch: Tensor, list, tuple or dict. Non-Tensor objects are ignored. Tensors are moved to ``device``.
         device: Device.
 
     Returns:
         Input with Tensors moved to device.
     """
-    if isinstance(batch, Tensor):
-        batch = batch.to(device)
-    elif isinstance(batch, dict):
-        batch = {k: to_device(b, device) for k, b in batch.items()}
-    elif isinstance(batch, (list, tuple)):
-        batch = type(batch)([to_device(b, device) for b in batch])
-    return batch
+    return tensor_to(batch, device)
 
 
 def asnumpy(v):
@@ -177,7 +201,13 @@ def fetch_model(name, map_location=None, **kwargs):
 
     """
     url = name if name.startswith('http') else f'https://celldetection.org/torch/models/{name}.pt'
-    return load_state_dict_from_url(url, map_location=map_location, **kwargs)
+    m = load_state_dict_from_url(url, map_location=map_location, **kwargs)
+    if isinstance(m, dict) and 'cd.models' in m.keys():
+        from .. import models
+        conf = m['cd.models']
+        m = getattr(models, conf['model'])(*conf['a'], **conf['kw'])
+        m.load_state_dict(conf['state_dict'])
+    return m
 
 
 def random_code_name(chars=4) -> str:
@@ -365,10 +395,18 @@ def train_epoch(model, train_loader, device, optimizer, desc=None, scaler=None,
         scheduler.step()
 
 
-def tweak_module_(module: nn.Module, class_or_tuple, must_exist=True, **kwargs):
+def iter_submodules(module: nn.Module, class_or_tuple, recursive=True):
+    for k, mod in module._modules.items():
+        if isinstance(mod, class_or_tuple):
+            yield module._modules, k, mod
+        if isinstance(mod, nn.Module) and recursive:
+            yield from iter_submodules(mod, class_or_tuple, recursive=recursive)
+
+
+def tweak_module_(module: nn.Module, class_or_tuple, must_exist=True, recursive=True, **kwargs):
     """Tweak module.
 
-    Sets attributes for all modules that are instances of given `class_or_tuple`.
+    Set attributes for all modules that are instances of given `class_or_tuple`.
 
     Examples:
         >>> import celldetection as cd, torch.nn as nn
@@ -382,41 +420,67 @@ def tweak_module_(module: nn.Module, class_or_tuple, must_exist=True, **kwargs):
         module: PyTorch `Module`.
         class_or_tuple: All instances of given `class_or_tuple` are to be tweaked.
         must_exist: If `True` an AttributeError is raised if keywords do not exist.
+        recursive: Whether to search for modules recursively.
         **kwargs: Attributes to be tweaked: `<attribute_name>=<value>`.
     """
-    for mod in module.modules():
-        if isinstance(mod, class_or_tuple):
-            for k, v in kwargs.items():
-                if must_exist:
-                    getattr(mod, k)
-                setattr(mod, k, v)
+    for handle, key, mod in iter_submodules(module, class_or_tuple, recursive=recursive):
+        for k, v in kwargs.items():
+            if must_exist:
+                getattr(mod, k)
+            setattr(mod, k, v)
 
 
-def replace_module_(module: nn.Module, class_or_tuple, substitute: nn.Module, recursive=True, **kwargs):
+def replace_module_(module: nn.Module, class_or_tuple, substitute: Union[Type[nn.Module], nn.Module], recursive=True,
+                    inherit_attr: list = None, **kwargs):
     """Replace module.
 
     Replace all occurrences of `class_or_tuple` in `module` with `substitute`.
 
     Examples:
         >>> # Replace all ReLU activations with LeakyReLU
-        ... replace_module_(network, nn.ReLU, nn.LeakyReLU)
+        ... cd.replace_module_(network, nn.ReLU, nn.LeakyReLU)
+
+        >>> # Replace all BatchNorm layers with InstanceNorm and inherit `num_features` attribute
+        ... cd.replace_module_(network, nn.BatchNorm2d, nn.InstanceNorm2d, inherit_attr=['num_features'])
 
     Args:
         module: Module.
         class_or_tuple: Class or tuple of classes that are to be replaced.
         substitute: Substitute class or object.
         recursive: Whether to replace modules recursively.
+        inherit_attr: Attributes to be inherited. List of attribute names. Attribute values are retrieved from replaced
+            module and passed to substitute constructor.
         **kwargs: Keyword arguments passed to substitute constructor if it is a class.
 
     """
-    for k, mod in module._modules.items():
-        if isinstance(mod, class_or_tuple):
-            if type(substitute) == type:
-                module._modules[k] = substitute(**kwargs)
-            else:
-                module._modules[k] = substitute
-        elif isinstance(mod, nn.Module) and recursive:
-            replace_module_(mod, substitute=substitute, class_or_tuple=class_or_tuple, recursive=recursive, **kwargs)
+    for handle, k, mod in iter_submodules(module, class_or_tuple, recursive=recursive):
+        if isinstance(substitute, nn.Module):
+            handle[k] = substitute
+        else:
+            inherit_attr = inherit_attr or []
+            handle[k] = substitute(**kwargs, **{k: mod.__dict__[k] for k in inherit_attr})
+
+
+def wrap_module_(module: nn.Module, class_or_tuple, wrapper, recursive=True, **kwargs):
+    for handle, k, mod in iter_submodules(module, class_or_tuple, recursive=recursive):
+        handle[k] = wrapper(handle[k], **kwargs)
+
+
+def spectral_norm_(module, class_or_tuple=nn.Conv2d, recursive=True, **kwargs):
+    """Spectral normalization.
+
+    Applies spectral normalization to parameters of all occurrences of ``class_or_tuple`` in the given module.
+
+    Note:
+        This is an inplace operation.
+
+    Args:
+        module: Module.
+        class_or_tuple: Class or tuple of classes whose parameters are to be normalized.
+        recursive: Whether to search for modules recursively.
+        **kwargs: Additional keyword arguments for ``torch.nn.utils.spectral_norm``.
+    """
+    wrap_module_(module, class_or_tuple, recursive=recursive, wrapper=nn.utils.spectral_norm, **kwargs)
 
 
 def get_device(module: Union[nn.Module, Tensor, torch.device]):
@@ -563,6 +627,7 @@ class Percent(float):
     Printable float that represents percentage.
 
     """
+
     def __str__(self):
         return '%g%%' % np.round(self, 2)
 
@@ -648,3 +713,120 @@ class Tiling:
             num_tiles=self.num_tiles,
             num_tiles_per_dim=self.num_tiles_per_dim
         )
+
+
+def to_h5(filename, mode='w', chunks=False, compression=None, overwrite=False, create_dataset_kw: dict = None,
+          **kwargs):
+    """To hdf5 file.
+
+    Write data to hdf5 file.
+
+    Args:
+        filename: File name.
+        mode: Mode.
+        chunks: Chunks setting for created datasets. Chunk shape, or True to enable auto-chunking.
+        compression: Compression setting for created datasets. Legal values are 'gzip', 'szip', 'lzf'. If an integer
+            in range(10), this indicates gzip compression level. Otherwise, an integer indicates the number of a
+            dynamically loaded compression filter.
+        overwrite: Whether to overwrite existing dataset.
+        create_dataset_kw: Additional keyword arguments for ``h5py.File().create_dataset``.
+        **kwargs: Data as ``{dataset_name: data}``.
+
+    """
+    create_dataset_kw = {} if create_dataset_kw is None else create_dataset_kw
+    with h5py.File(filename, mode) as h:
+        for k, v in kwargs.items():
+            exists = k in h
+            if overwrite and exists:
+                del h[k]
+            elif exists:
+                h[k][:] = v
+            else:
+                h.create_dataset(k, data=v, compression=compression, chunks=chunks, **create_dataset_kw)
+
+
+def to_tiff(filename, image, mode='w', method='tile', bigtiff=True):
+    """To tiff file.
+
+    Write ``image`` to tiff file using ``pytiff``.
+    By default, the tiff is tiled, s.t. crops can be read from disk without loading the entire image into memory first.
+
+    Notes:
+        - ``pytiff`` must be installed to use this function.
+
+    References:
+        https://pytiff.readthedocs.io/en/master/quickstart.html
+
+    Args:
+        filename: File name.
+        image: Image.
+        mode: Mode.
+        method: Method. Either ``'tile'`` or ``'scanline'``.
+        bigtiff: Whether to use bigtiff format.
+
+    """
+    try:
+        from pytiff import Tiff
+    except ModuleNotFoundError:
+        raise ModuleNotFoundError('To use the to_tiff function pytiff must be installed.\n'
+                                  'See: https://pytiff.readthedocs.io/en/master/quickstart.html')
+    with Tiff(filename, mode, bigtiff=bigtiff) as handle:
+        handle.write(image, method=method)
+
+
+def exponential_moving_average_(module_avg, module, alpha=.999, alpha_non_trainable=0., buffers=True):
+    """Exponential moving average.
+
+    Update the variables of ``module_avg`` to be slightly closer to ``module``.
+
+    References:
+        - https://arxiv.org/pdf/1710.10196.pdf
+        - https://arxiv.org/pdf/2006.07733.pdf
+
+    Notes:
+        - Whether a parameter is trainable or not is checked on ``module``
+        - ``module_avg`` can be on different device and entirely frozen
+
+    Args:
+        module_avg: Average module. The parameters of this model are to be updated.
+        module: Other Module.
+        alpha: Fraction of trainable parameters of ``module_avg``; (1 - alpha) is fraction of trainable
+            parameters of ``module``.
+        alpha_non_trainable: Same as ``alpha``, but for non-trainable parameters.
+        buffers: Whether to copy buffers from ``module`` to ``module_avg``.
+    """
+    device = get_device(module_avg)
+    with torch.no_grad():
+        for avg, new in zip(_params(module_avg), _params(module)):
+            a = alpha if new.requires_grad else alpha_non_trainable
+            avg.data.mul_(a).add_(new.data.to(device), alpha=1 - a)
+    if buffers:
+        for avg, new in zip(module_avg.buffers(), module.buffers()):
+            avg.copy_(new)
+
+
+def to_json(filename, obj, mode='w'):
+    """To JSON.
+
+    Dump ``obj`` to JSON file with name ``filename``.
+
+    Args:
+        filename: File name.
+        obj: Object.
+        mode: File mode.
+    """
+    with open(filename, mode) as fp:
+        json.dump(obj, fp)
+
+
+def from_json(filename):
+    """From JSON.
+
+    Load object from JSON file with name ``filename``.
+
+    Args:
+        filename: File name.
+    """
+    with open(filename, 'r') as fp:
+        v = json.load(fp)
+    return v
